@@ -1,6 +1,6 @@
 /*
  * Copyright © 2009 CNRS
- * Copyright © 2009-2021 Inria.  All rights reserved.
+ * Copyright © 2009-2022 Inria.  All rights reserved.
  * Copyright © 2009-2011 Université Bordeaux
  * Copyright © 2011 Cisco Systems, Inc.  All rights reserved.
  * Copyright © 2011      Oracle and/or its affiliates.  All rights reserved.
@@ -493,6 +493,9 @@ hwloc_look_lgrp(struct hwloc_topology *topology, struct hwloc_disc_status *dstat
     dstatus->flags |= HWLOC_DISC_STATUS_FLAG_GOT_ALLOWED_RESOURCES;
   }
 
+  if (topology->flags & HWLOC_TOPOLOGY_FLAG_NO_DISTANCES)
+    need_distances = 0;
+
   cookie = lgrp_init(LGRP_VIEW_OS);
   if (cookie == LGRP_COOKIE_NONE)
     {
@@ -639,7 +642,7 @@ hwloc_look_kstat(struct hwloc_topology *topology)
 
       if (kstat_read(kc, ksp, NULL) == -1)
 	{
-          if (hwloc_hide_errors() < 2)
+          if (HWLOC_SHOW_CRITICAL_ERRORS())
             fprintf(stderr, "hwloc/solaris: kstat_read failed for CPU%u: %s\n", cpuid, strerror(errno));
 	  continue;
 	}
@@ -698,7 +701,7 @@ hwloc_look_kstat(struct hwloc_topology *topology)
 	if (!stat)
 	  {
 	    if (Lpkg_num) {
-              if (hwloc_hide_errors() < 2)
+              if (HWLOC_SHOW_CRITICAL_ERRORS())
                 fprintf(stderr, "hwloc/solaris: could not read package id for CPU%u: %s\n", cpuid, strerror(errno));
             } else {
 	      hwloc_debug("could not read package id for CPU%u: %s\n", cpuid, strerror(errno));
@@ -722,7 +725,7 @@ hwloc_look_kstat(struct hwloc_topology *topology)
 	    break;
 #endif
 	  default:
-            if (hwloc_hide_errors() < 2)
+            if (HWLOC_SHOW_CRITICAL_ERRORS())
               fprintf(stderr, "hwloc/solaris: chip_id type %u unknown\n", (unsigned) stat->data_type);
 	    look_chips = 0;
 	    continue;
@@ -751,7 +754,7 @@ hwloc_look_kstat(struct hwloc_topology *topology)
 	if (!stat)
 	  {
 	    if (Lcore_num) {
-              if (hwloc_hide_errors() < 2)
+              if (HWLOC_SHOW_CRITICAL_ERRORS())
                 fprintf(stderr, "hwloc/solaris: could not read core id for CPU%u: %s\n", cpuid, strerror(errno));
             } else {
 	      hwloc_debug("could not read core id for CPU%u: %s\n", cpuid, strerror(errno));
@@ -775,7 +778,7 @@ hwloc_look_kstat(struct hwloc_topology *topology)
 	    break;
 #endif
 	  default:
-            if (hwloc_hide_errors() < 2)
+            if (HWLOC_SHOW_CRITICAL_ERRORS())
               fprintf(stderr, "hwloc/solaris: core_id type %u unknown\n", (unsigned) stat->data_type);
 	    look_cores = 0;
 	    continue;
@@ -804,7 +807,7 @@ hwloc_look_kstat(struct hwloc_topology *topology)
 
     } else if (!strcmp("pg_hw_perf", ksp->ks_module)) {
       if (kstat_read(kc, ksp, NULL) == -1) {
-        if (hwloc_hide_errors() < 2)
+        if (HWLOC_SHOW_CRITICAL_ERRORS())
           fprintf(stderr, "hwloc/solaris: kstat_read failed for module %s name %s instance %d: %s\n", ksp->ks_module, ksp->ks_name, ksp->ks_instance, strerror(errno));
 	continue;
       }
