@@ -4,6 +4,7 @@
  * Copyright © 2009-2012, 2015, 2017 Université Bordeaux
  * Copyright © 2009-2011 Cisco Systems, Inc.  All rights reserved.
  * Copyright © 2020 Hewlett Packard Enterprise.  All rights reserved.
+ * Copyright © 2023 Université de Reims Champagne-Ardenne.  All rights reserved.
  * See COPYING in top-level directory.
  */
 
@@ -64,16 +65,11 @@ extern void setJNIEnv();
 
 FILE *open_output(const char *filename, int overwrite)
 {
-  const char *extn;
   struct stat st;
 
   if (!filename || !strcmp(filename, "-"))
     return stdout;
 
-  extn = strrchr(filename, '.');
-  if (filename[0] == '-' && extn == filename + 1)
-    return stdout;
-
   if (!stat(filename, &st) && !overwrite) {
     errno = EEXIST;
     return NULL;
@@ -82,7 +78,7 @@ FILE *open_output(const char *filename, int overwrite)
   return fopen(filename, "w");
 }
 
-static hwloc_obj_t insert_task(hwloc_topology_t topology, hwloc_cpuset_t cpuset, const char * name, int thread)
+static hwloc_obj_t insert_misc(hwloc_topology_t topology, hwloc_cpuset_t cpuset, const char *subtype, const char * name)
 {
   hwloc_obj_t group, obj;
 
@@ -103,16 +99,16 @@ static hwloc_obj_t insert_task(hwloc_topology_t topology, hwloc_cpuset_t cpuset,
     hwloc_bitmap_asprintf(&s, cpuset);
     group = hwloc_get_obj_covering_cpuset(topology, cpuset);
     hwloc_bitmap_asprintf(&gs, group->cpuset);
-    fprintf(stderr, "%s `%s' binding %s doesn't match any object, extended to %s before inserting the %s in the topology.\n",
-	    thread ? "Thread" : "Process", name, s, gs, thread ? "thread" : "process");
+    fprintf(stderr, "%s `%s' binding %s doesn't match any object, extended to %s before inserting the object.\n",
+	    subtype, name, s, gs);
     free(s);
     free(gs);
   }
   obj = hwloc_topology_insert_misc_object(topology, group, name);
   if (!obj)
     fprintf(stderr, "Failed to insert process `%s'\n", name);
-  else
-    obj->subtype = strdup("Process");
+  else if (subtype)
+    obj->subtype = strdup(subtype);
 
   return obj;
 }
@@ -121,6 +117,7 @@ static void foreach_process_cb(hwloc_topology_t topology,
 			       struct hwloc_ps_process *proc,
 			       void *cbdata __hwloc_attribute_unused)
 {
+  /* sort of similar to print_process_lstopo_misc() in hwloc-ps.c */
   char name[100];
   unsigned i;
 
@@ -129,7 +126,7 @@ static void foreach_process_cb(hwloc_topology_t topology,
     snprintf(name, sizeof(name), "%ld %s", proc->pid, proc->name);
 
   if (proc->bound)
-    insert_task(topology, proc->cpuset, name, 0);
+    insert_misc(topology, proc->cpuset, "Process", name);
 
   if (proc->nthreads)
     for(i=0; i<proc->nthreads; i++)
@@ -141,7 +138,7 @@ static void foreach_process_cb(hwloc_topology_t topology,
 	else
 	  snprintf(task_name, sizeof(task_name), "%s %li", name, proc->threads[i].tid);
 
-	insert_task(topology, proc->threads[i].cpuset, task_name, 1);
+	insert_misc(topology, proc->threads[i].cpuset, "Thread", task_name);
       }
 }
 
@@ -158,6 +155,77 @@ static void add_process_objects(hwloc_topology_t topology)
 			   HWLOC_PS_FLAG_THREADS | HWLOC_PS_FLAG_SHORTNAME, NULL, HWLOC_PS_ALL_UIDS);
 }
 
+static void add_one_misc_object_from(hwloc_topology_t topology,
+                                     char *subtype, char *name, hwloc_bitmap_t cpuset)
+{
+  if (!hwloc_bitmap_iszero(cpuset) && subtype && name) {
+    insert_misc(topology, cpuset, subtype, name);
+  } else {
+    char *s;
+    hwloc_bitmap_asprintf(&s, cpuset);
+    fprintf(stderr, "Ignoring misc object subtype %s name %s cpuset %s\n", subtype, name, s);
+    free(s);
+  }
+}
+
+/* reads Misc description from the FILE*
+ * entries must look like:
+ * name=... (must be first)
+ * cpuset=... (cannot be 0)
+ * subtype=... (optional)
+ */
+static void add_misc_objects_from(hwloc_topology_t topology, FILE *from)
+{
+  char line[256];
+  hwloc_bitmap_t cpuset;
+  char *subtype = NULL;
+  char *name = NULL;
+  cpuset = hwloc_bitmap_alloc();
+  if (!cpuset)
+    return;
+
+  while (fgets(line, sizeof line, from)) {
+    char *end;
+
+    /* remove ending \n */
+    end = strchr(line, '\n');
+    if (end)
+      *end = '\0';
+    /* ignoring empty lines */
+    if (line[0] == '\0')
+      continue;
+
+    if (!strncmp(line, "name=", 5)) {
+      /* commit (or ignore) the previous entry */
+      if (name)
+        add_one_misc_object_from(topology, subtype, name, cpuset);
+      /* start a new entry */
+      free(subtype);
+      subtype = NULL;
+      free(name);
+      name = strdup(line+5);
+      hwloc_bitmap_zero(cpuset);
+
+    } else if (!strncmp(line, "cpuset=", 7)) {
+      hwloc_bitmap_sscanf(cpuset, line+7);
+
+    } else if (!strncmp(line, "subtype=", 8)) {
+      free(subtype);
+      subtype = strdup(line+8);
+
+    } else {
+      fprintf(stderr, "Unrecognized --misc-from line `%s', ignored\n", line);
+    }
+  }
+
+  /* commit (or ignore) the last entry */
+  if (name)
+    add_one_misc_object_from(topology, subtype, name, cpuset);
+  free(name);
+  free(subtype);
+  hwloc_bitmap_free(cpuset);
+}
+
 static __hwloc_inline void lstopo_update_factorize_bounds(unsigned min, unsigned *first, unsigned *last)
 {
   switch (min) {
@@ -308,11 +376,13 @@ lstopo_check_pci_domains(hwloc_topology_t topology)
 
 static void
 lstopo_parse_children_order(char *s, unsigned *children_order_p,
+                            enum lstopo_orient_e *above_force_orient_p,
                             enum lstopo_orient_e *right_force_orient_p,
                             enum lstopo_orient_e *below_force_orient_p)
 {
   char *tmp, *next;
   unsigned children_order;
+  enum lstopo_orient_e above_force_orient = LSTOPO_ORIENT_NONE;
   enum lstopo_orient_e right_force_orient = LSTOPO_ORIENT_NONE;
   enum lstopo_orient_e below_force_orient= LSTOPO_ORIENT_NONE;
 
@@ -332,6 +402,15 @@ lstopo_parse_children_order(char *s, unsigned *children_order_p,
 
     if (!strcmp(tmp, "memory:above") || !strcmp(tmp, "memoryabove") /* backward compat with 2.5 */) {
       children_order |= LSTOPO_ORDER_MEMORY_ABOVE;
+    } else if (!strcmp(tmp, "memory:above:horiz")) {
+      children_order |= LSTOPO_ORDER_MEMORY_ABOVE;
+      above_force_orient = LSTOPO_ORIENT_HORIZ;
+    } else if (!strcmp(tmp, "memory:above:vert")) {
+      children_order |= LSTOPO_ORDER_MEMORY_ABOVE;
+      above_force_orient = LSTOPO_ORIENT_VERT;
+    } else if (!strcmp(tmp, "memory:above:rect")) {
+      children_order |= LSTOPO_ORDER_MEMORY_ABOVE;
+      above_force_orient = LSTOPO_ORIENT_RECT;
 
     } else if (!strcmp(tmp, "io:right")) {
       children_order |= LSTOPO_ORDER_IO_RIGHT;
@@ -389,6 +468,7 @@ lstopo_parse_children_order(char *s, unsigned *children_order_p,
   }
 
   *children_order_p = children_order;
+  *above_force_orient_p = above_force_orient;
   *right_force_orient_p = right_force_orient;
   *below_force_orient_p = below_force_orient;
 }
@@ -482,7 +562,7 @@ void usage(const char *name, FILE *where)
   fprintf (where, "Textual output options:\n");
   fprintf (where, "  --only <type>         Only show objects of the given type in the textual output\n");
   fprintf (where, "  -v --verbose          Include additional details\n");
-  fprintf (where, "  -s --silent           Reduce the amount of details to show\n");
+  fprintf (where, "  -q --quiet -s         Reduce the amount of details to show\n");
   fprintf (where, "  --distances           Only show distance matrices\n");
   fprintf (where, "  --distances-transform <links|merge-switch-ports|transitive-closure>\n");
   fprintf (where, "                        Transform distances before displaying them\n");
@@ -567,6 +647,7 @@ void usage(const char *name, FILE *where)
 		  "                        Set flags during the synthetic topology export\n");
   /* --shmem-output-addr is undocumented on purpose */
   fprintf (where, "  --ps --top            Display processes within the hierarchy\n");
+  fprintf (where, "  --misc-from <file>    Create Misc objects as defined in <file>");
   fprintf (where, "  --version             Report version and exit\n");
   fprintf (where, "  -h --help             Show this usage\n");
 }
@@ -712,6 +793,30 @@ parse_output_format(const char *name, char *callname __hwloc_attribute_unused)
     return LSTOPO_OUTPUT_ERROR;
 }
 
+static const char *
+output_format_name(enum output_format format)
+{
+  switch (format) {
+  case LSTOPO_OUTPUT_DEFAULT: return "default";
+  case LSTOPO_OUTPUT_WINDOW: return "window";
+  case LSTOPO_OUTPUT_CONSOLE: return "console";
+  case LSTOPO_OUTPUT_SYNTHETIC: return "synthetic";
+  case LSTOPO_OUTPUT_ASCII: return "ascii";
+  case LSTOPO_OUTPUT_TIKZ: return "tikz";
+  case LSTOPO_OUTPUT_FIG: return "fig";
+  case LSTOPO_OUTPUT_PNG: return "png";
+  case LSTOPO_OUTPUT_PDF: return "pdf";
+  case LSTOPO_OUTPUT_PS: return "ps";
+  case LSTOPO_OUTPUT_SVG: return "svg";
+  case LSTOPO_OUTPUT_CAIROSVG: return "cairosvg";
+  case LSTOPO_OUTPUT_NATIVESVG: return "nativesvg";
+  case LSTOPO_OUTPUT_XML: return "xml";
+  case LSTOPO_OUTPUT_SHMEM: return "shmem";
+  case LSTOPO_OUTPUT_ERROR: return "error";
+  }
+  abort();
+}
+
 /****************************************************
  * Store filters during parsing and apply them later
  */
@@ -783,7 +888,8 @@ main (int argc, char *argv[])
   hwloc_bitmap_t allow_cpuset = NULL, allow_nodeset = NULL;
   char * callname;
   char * input = NULL;
-  enum hwloc_utils_input_format input_format = HWLOC_UTILS_INPUT_DEFAULT;
+  const char *show_only_string = NULL;
+  struct hwloc_utils_input_format_s input_format = HWLOC_UTILS_INPUT_FORMAT_DEFAULT;
   enum output_format output_format = LSTOPO_OUTPUT_DEFAULT;
   struct lstopo_type_filter type_filter[HWLOC_OBJ_TYPE_MAX];
   char *restrictstring = NULL;
@@ -797,6 +903,7 @@ main (int argc, char *argv[])
 #endif
   char *env;
   int top = 0;
+  FILE * miscfrom = NULL;
   int opt;
   unsigned i;
 
@@ -844,7 +951,6 @@ main (int argc, char *argv[])
   loutput.show_memattrs_only = 0;
   loutput.show_cpukinds_only = 0;
   loutput.show_windows_processor_groups_only = 0;
-  loutput.show_only = HWLOC_OBJ_TYPE_NONE;
   loutput.show_cpuset = 0;
   loutput.show_taskset = 0;
   loutput.transform_distances = -1;
@@ -873,6 +979,7 @@ main (int argc, char *argv[])
     loutput.force_orient[i] = LSTOPO_ORIENT_HORIZ;
   loutput.force_orient[HWLOC_OBJ_NUMANODE] = LSTOPO_ORIENT_HORIZ;
   loutput.force_orient[HWLOC_OBJ_MEMCACHE] = LSTOPO_ORIENT_HORIZ;
+  loutput.above_force_orient = LSTOPO_ORIENT_NONE;
   loutput.right_force_orient = LSTOPO_ORIENT_NONE;
   loutput.below_force_orient = LSTOPO_ORIENT_NONE;
   for(i=HWLOC_OBJ_TYPE_MIN; i<HWLOC_OBJ_TYPE_MAX; i++) {
@@ -916,7 +1023,8 @@ main (int argc, char *argv[])
       opt = 0;
       if (!strcmp (argv[0], "-v") || !strcmp (argv[0], "--verbose")) {
 	loutput.verbose_mode++;
-      } else if (!strcmp (argv[0], "-s") || !strcmp (argv[0], "--silent")) {
+      } else if (!strcmp (argv[0], "-q") || !strcmp (argv[0], "--quiet")
+                 || !strcmp (argv[0], "-s") || !strcmp (argv[0], "--silent")) {
 	loutput.verbose_mode--;
       } else if (!strcmp (argv[0], "--distances")) {
 	loutput.show_distances_only = 1;
@@ -962,8 +1070,7 @@ main (int argc, char *argv[])
       } else if (!strcmp (argv[0], "--only")) {
 	if (argc < 2)
 	  goto out_usagefailure;
-        if (hwloc_type_sscanf(argv[1], &loutput.show_only, NULL, 0) < 0)
-	  fprintf(stderr, "Unsupported type `%s' passed to --only, ignoring.\n", argv[1]);
+        show_only_string = argv[1];
 	opt = 1;
       }
       else if (!strcmp (argv[0], "--filter")) {
@@ -1354,7 +1461,7 @@ main (int argc, char *argv[])
 	if (argc < 2)
 	  goto out_usagefailure;
         lstopo_parse_children_order(argv[1], &loutput.children_order,
-                                    &loutput.right_force_orient, &loutput.below_force_orient);
+                                    &loutput.above_force_orient, &loutput.right_force_orient, &loutput.below_force_orient);
 	opt = 1;
       }
       else if (!strcmp (argv[0], "--no-cpukinds")) {
@@ -1423,7 +1530,19 @@ main (int argc, char *argv[])
 	loutput.pid_number = atoi(argv[1]); opt = 1;
       } else if (!strcmp (argv[0], "--ps") || !strcmp (argv[0], "--top"))
         top = 1;
-      else if (!strcmp (argv[0], "--version")) {
+      else if (!strcmp (argv[0], "--misc-from")) {
+	if (argc < 2)
+	  goto out_usagefailure;
+        if (!strcmp(argv[1], "-"))
+          miscfrom = stdin;
+        else
+          miscfrom = fopen(argv[1], "r");
+        if (!miscfrom) {
+          fprintf(stderr, "Failed open --misc-from %s file for reading (%s)\n", argv[1], strerror(errno));
+          exit(EXIT_FAILURE);
+        }
+        opt = 1;
+      } else if (!strcmp (argv[0], "--version")) {
           printf("%s %s\n", callname, HWLOC_VERSION);
           exit(EXIT_SUCCESS);
       } else if (!strcmp (argv[0], "--output-format") || !strcmp (argv[0], "--of")) {
@@ -1453,16 +1572,27 @@ main (int argc, char *argv[])
    * Configure the output
    */
 
+  /* if output is -.format but --of was given, ignore .format */
+  if (output_format != LSTOPO_OUTPUT_DEFAULT
+      && filename && filename[0] == '-' && filename[1] == '.') {
+    fprintf(stderr, "Ignoring extension in stdout output `%s' since --of was also given format `%s'.\n",
+            filename, output_format_name(output_format));
+    filename = "-"; /* to simplify things later */
+  }
+
   /* if the output format wasn't enforced, look at the filename */
   if (filename && output_format == LSTOPO_OUTPUT_DEFAULT) {
     if (!strcmp(filename, "-")
 	|| !strcmp(filename, "/dev/stdout")) {
       output_format = LSTOPO_OUTPUT_CONSOLE;
+      filename = "-"; /* to simplify things later */
     } else {
       char *dot = strrchr(filename, '.');
-      if (dot)
+      if (dot) {
         output_format = parse_output_format(dot+1, callname);
-      else {
+        if (dot == filename+1 && filename[0] == '-' && output_format != LSTOPO_OUTPUT_ERROR)
+          filename = "-"; /* to simplify things later */
+      } else {
 	fprintf(stderr, "Cannot infer output type for file `%s' without any extension, using default output.\n", filename);
 	filename = NULL;
       }
@@ -1474,7 +1604,7 @@ main (int argc, char *argv[])
   /* if  the output format wasn't enforced, think a bit about what the user probably want */
   if (output_format == LSTOPO_OUTPUT_DEFAULT) {
     if (loutput.show_cpuset
-        || loutput.show_only != HWLOC_OBJ_TYPE_NONE
+        || show_only_string
 	|| loutput.show_distances_only
         || loutput.show_memattrs_only
         || loutput.show_cpukinds_only
@@ -1500,22 +1630,26 @@ main (int argc, char *argv[])
 #endif
     if (getenv("DISPLAY") && !want_console) {
       output_func = output_x11;
+      output_format = LSTOPO_OUTPUT_WINDOW;
     } else
 #endif /* LSTOPO_HAVE_X11 */
 #ifdef HWLOC_WIN_SYS
     {
       output_func = output_windows;
+      output_format = LSTOPO_OUTPUT_WINDOW;
     }
 #endif
 #endif /* !LSTOPO_HAVE_GRAPHICS */
 #if !defined HWLOC_WIN_SYS || !defined LSTOPO_HAVE_GRAPHICS
     {
       output_func = output_console;
+      output_format = LSTOPO_OUTPUT_CONSOLE;
     }
 #endif
 #ifdef ANDROID
     setJNIEnv();
     output_func = output_android;
+    output_format = LSTOPO_OUTPUT_WINDOW;
 #endif
     }
     break;
@@ -1571,6 +1705,7 @@ main (int argc, char *argv[])
   case LSTOPO_OUTPUT_SVG:
   case LSTOPO_OUTPUT_CAIROSVG:
     output_func = output_cairosvg;
+    output_format = LSTOPO_OUTPUT_CAIROSVG;
     break;
 # endif /* CAIRO_HAS_SVG_SURFACE */
 #endif /* LSTOPO_HAVE_GRAPHICS */
@@ -1579,6 +1714,7 @@ main (int argc, char *argv[])
 #endif
   case LSTOPO_OUTPUT_NATIVESVG:
     output_func = output_nativesvg;
+    output_format = LSTOPO_OUTPUT_NATIVESVG;
     break;
   case LSTOPO_OUTPUT_XML:
     output_func = output_xml;
@@ -1617,7 +1753,7 @@ main (int argc, char *argv[])
     if (err)
       goto out_with_topology;
 
-    if (input_format != HWLOC_UTILS_INPUT_DEFAULT) {
+    if (input_format.format != HWLOC_UTILS_INPUT_DEFAULT) {
       /* add the input path to the window title */
       snprintf(loutput.title, sizeof(loutput.title), "lstopo - %s", input);
 
@@ -1650,7 +1786,7 @@ main (int argc, char *argv[])
     }
   }
 
-  if (input_format == HWLOC_UTILS_INPUT_XML
+  if (input_format.format == HWLOC_UTILS_INPUT_XML
       && output_format == LSTOPO_OUTPUT_XML) {
     /* must be after parsing output format and before loading the topology */
     putenv((char *) "HWLOC_XML_USERDATA_NOT_DECODED=1");
@@ -1669,7 +1805,7 @@ main (int argc, char *argv[])
     clock_gettime(CLOCK_MONOTONIC, &ts1);
 #endif
 
-  if (input_format == HWLOC_UTILS_INPUT_SHMEM) {
+  if (input_format.format == HWLOC_UTILS_INPUT_SHMEM) {
 #ifdef HWLOC_WIN_SYS
     fprintf(stderr, "shmem topology not supported\n"); /* this line must match the grep line in test-lstopo-shmem */
     goto out_with_topology;
@@ -1699,6 +1835,14 @@ main (int argc, char *argv[])
   }
 #endif
 
+  /********************************
+   * Clean input format's internal variables
+   */
+
+  if (input) {
+    hwloc_utils_disable_input_format(&input_format);
+  }
+
   /********************************
    * Tweak the topology and output
    */
@@ -1733,6 +1877,8 @@ main (int argc, char *argv[])
 
   if (top)
     add_process_objects(topology);
+  if (miscfrom)
+    add_misc_objects_from(topology, miscfrom);
 
   if (restrictstring) {
     hwloc_bitmap_t restrictset = hwloc_bitmap_alloc();
@@ -1754,6 +1900,20 @@ main (int argc, char *argv[])
   loutput.depth = hwloc_topology_get_depth(topology);
   loutput.file = NULL;
 
+  if (filename && output_format == LSTOPO_OUTPUT_WINDOW) {
+    fprintf(stderr, "Output filename `%s' ignored when using graphical window output.\n",
+            filename);
+  }
+
+#ifdef HAVE_ISATTY
+  /* if exporting to a file (which is not stdout, rewritten as "-" above),
+   * and stdout is a terminal,
+   * show a message about the export filename and format.
+   */
+  if (filename && strcmp(filename, "-") && output_format != LSTOPO_OUTPUT_WINDOW && isatty(STDOUT_FILENO))
+    printf("Exporting format `%s' to file `%s'\n", output_format_name(output_format), filename);
+#endif
+
   if (output_format != LSTOPO_OUTPUT_XML) {
     /* there might be some xml-imported userdata in objects, add lstopo-specific userdata in front of them */
     lstopo_populate_userdata(hwloc_get_root_obj(topology));
@@ -1763,6 +1923,16 @@ main (int argc, char *argv[])
     lstopo_add_collapse_attributes(topology);
   }
 
+  loutput.show_only.depth = HWLOC_TYPE_DEPTH_UNKNOWN; /* disable this feature by default */
+  if (show_only_string) {
+    err = hwloc_calc_parse_level(NULL, topology, show_only_string, strlen(show_only_string), &loutput.show_only);
+    if (err < 0 && loutput.show_only.depth == HWLOC_TYPE_DEPTH_UNKNOWN) {
+      fprintf(stderr, "level %s passed to --only is unavailable.\n", show_only_string);
+      goto out_with_topology;
+    }
+    /* multiple depth is ok */
+  }
+
   /******************
    * Output for real
    */
@@ -1799,9 +1969,12 @@ main (int argc, char *argv[])
   lstopo_destroy_userdata(hwloc_get_root_obj(topology));
   hwloc_topology_destroy(topology);
  out:
+  if (input) hwloc_utils_disable_input_format(&input_format);
   hwloc_bitmap_free(allow_cpuset);
   hwloc_bitmap_free(allow_nodeset);
   hwloc_bitmap_free(loutput.cpubind_set);
   hwloc_bitmap_free(loutput.membind_set);
+  if (miscfrom && miscfrom != stdin)
+    fclose(miscfrom);
   return EXIT_FAILURE;
 }
